@@ -10,9 +10,11 @@
 
 //! Virtual machine and JIT compiler for eBPF programs.
 
+use crate::disassembler::disassemble_instruction;
+use crate::static_analysis::Analysis;
 use crate::{
     call_frames::CallFrames,
-    disassembler, ebpf,
+    ebpf,
     elf::EBpfElf,
     error::{EbpfError, UserDefinedError},
     jit::{JitProgram, JitProgramArgument},
@@ -20,7 +22,11 @@ use crate::{
     user_error::UserError,
 };
 use log::debug;
-use std::{collections::HashMap, fmt::Debug, u32};
+use std::{
+    collections::{BTreeMap, HashMap},
+    fmt::Debug,
+    u32,
+};
 
 /// eBPF verification function that returns an error if the program does not meet its requirements.
 ///
@@ -198,10 +204,8 @@ pub trait Executable<E: UserDefinedError, I: InstructionMeter>: Send + Sync {
     fn get_ro_sections(&self) -> Result<Vec<(u64, &[u8])>, EbpfError<E>>;
     /// Get the entry point offset into the text section
     fn get_entrypoint_instruction_offset(&self) -> Result<usize, EbpfError<E>>;
-    /// Set a symbol's instruction offset
-    fn register_bpf_function(&mut self, hash: u32, pc: usize);
     /// Get a symbol's instruction offset
-    fn lookup_bpf_function(&self, hash: u32) -> Option<&usize>;
+    fn lookup_bpf_function(&self, hash: u32) -> Option<usize>;
     /// Get the syscall registry
     fn get_syscall_registry(&self) -> &SyscallRegistry;
     /// Set (overwrite) the syscall registry
@@ -213,7 +217,7 @@ pub trait Executable<E: UserDefinedError, I: InstructionMeter>: Send + Sync {
     /// Report information on a symbol that failed to be resolved
     fn report_unresolved_symbol(&self, insn_offset: usize) -> Result<u64, EbpfError<E>>;
     /// Get syscalls and BPF functions (if debug symbols are not stripped)
-    fn get_symbols(&self) -> (HashMap<u32, String>, HashMap<usize, (String, usize)>);
+    fn get_symbols(&self) -> (BTreeMap<u32, String>, BTreeMap<usize, (u32, String)>);
 }
 
 /// Static constructors for Executable
@@ -234,13 +238,18 @@ impl<E: UserDefinedError, I: 'static + InstructionMeter> dyn Executable<E, I> {
     /// Creates a post relocaiton/fixup executable from machine code
     pub fn from_text_bytes(
         text_bytes: &[u8],
+        bpf_functions: BTreeMap<u32, (usize, String)>,
         verifier: Option<Verifier<E>>,
         config: Config,
     ) -> Result<Box<Self>, EbpfError<E>> {
         if let Some(verifier) = verifier {
             verifier(text_bytes)?;
         }
-        Ok(Box::new(EBpfElf::new_from_text_bytes(config, text_bytes)))
+        Ok(Box::new(EBpfElf::new_from_text_bytes(
+            config,
+            text_bytes,
+            bpf_functions,
+        )))
     }
 }
 
@@ -262,6 +271,43 @@ impl InstructionMeter for DefaultInstructionMeter {
     }
 }
 
+/// Statistic of taken branches (from a recorded trace)
+pub struct DynamicAnalysis {
+    /// Maximal edge counter value
+    pub edge_counter_max: usize,
+    /// src_node, dst_node, edge_counter
+    pub edges: BTreeMap<usize, BTreeMap<usize, usize>>,
+}
+
+impl DynamicAnalysis {
+    /// Accumulates a trace
+    pub fn new<E: UserDefinedError, I: InstructionMeter>(
+        tracer: &Tracer,
+        analysis: &Analysis<E, I>,
+    ) -> Self {
+        let mut result = Self {
+            edge_counter_max: 0,
+            edges: BTreeMap::new(),
+        };
+        let mut last_basic_block = std::usize::MAX;
+        for traced_instruction in tracer.log.iter() {
+            let pc = traced_instruction[11] as usize;
+            if analysis.cfg_nodes.contains_key(&pc) {
+                let counter = result
+                    .edges
+                    .entry(last_basic_block)
+                    .or_insert_with(BTreeMap::new)
+                    .entry(pc)
+                    .or_insert(0);
+                *counter += 1;
+                result.edge_counter_max = result.edge_counter_max.max(*counter);
+                last_basic_block = pc;
+            }
+        }
+        result
+    }
+}
+
 /// Used for instruction tracing
 #[derive(Default, Clone)]
 pub struct Tracer {
@@ -276,33 +322,41 @@ impl Tracer {
     }
 
     /// Use this method to print the log of this tracer
-    pub fn write<W: std::fmt::Write>(
+    pub fn write<W: std::io::Write, E: UserDefinedError, I: InstructionMeter>(
         &self,
-        out: &mut W,
-        program: &[u8],
-    ) -> Result<(), std::fmt::Error> {
-        let disassembled = disassembler::to_insn_vec(program);
-        let mut pc_to_instruction_index =
-            vec![0usize; disassembled.last().map(|ins| ins.ptr + 2).unwrap_or(0)];
-        for index in 0..disassembled.len() {
-            pc_to_instruction_index[disassembled[index].ptr] = index;
-            pc_to_instruction_index[disassembled[index].ptr + 1] = index;
+        output: &mut W,
+        analysis: &Analysis<E, I>,
+    ) -> Result<(), std::io::Error> {
+        let mut pc_to_insn_index = vec![
+            0usize;
+            analysis
+                .instructions
+                .last()
+                .map(|insn| insn.ptr + 2)
+                .unwrap_or(0)
+        ];
+        for (index, insn) in analysis.instructions.iter().enumerate() {
+            pc_to_insn_index[insn.ptr] = index;
+            pc_to_insn_index[insn.ptr + 1] = index;
         }
         for index in 0..self.log.len() {
             let entry = &self.log[index];
+            let pc = entry[11] as usize;
+            let insn = &analysis.instructions[pc_to_insn_index[pc]];
             writeln!(
-                out,
+                output,
                 "{:5?} {:016X?} {:5?}: {}",
                 index,
                 &entry[0..11],
-                entry[11] as usize + ebpf::ELF_INSN_DUMP_OFFSET,
-                disassembled[pc_to_instruction_index[entry[11] as usize]].desc,
+                pc + ebpf::ELF_INSN_DUMP_OFFSET,
+                disassemble_instruction(&insn, &analysis),
             )?;
         }
         Ok(())
     }
 
     /// Compares an interpreter trace and a JIT trace.
+    ///
     /// The log of the JIT can be longer because it only validates the instruction meter at branches.
     pub fn compare(interpreter: &Self, jit: &Self) -> bool {
         let interpreter = interpreter.log.as_slice();
@@ -354,7 +408,7 @@ pub const SYSCALL_CONTEXT_OBJECTS_OFFSET: usize = 6;
 /// # Examples
 ///
 /// ```
-/// use solana_rbpf::{vm::{Config, Executable, EbpfVm, DefaultInstructionMeter}, user_error::UserError};
+/// use solana_rbpf::{ebpf, elf::register_bpf_function, vm::{Config, Executable, EbpfVm, DefaultInstructionMeter}, user_error::UserError};
 ///
 /// let prog = &[
 ///     0x95, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00  // exit
@@ -364,7 +418,9 @@ pub const SYSCALL_CONTEXT_OBJECTS_OFFSET: usize = 6;
 /// ];
 ///
 /// // Instantiate a VM.
-/// let executable = Executable::<UserError, DefaultInstructionMeter>::from_text_bytes(prog, None, Config::default()).unwrap();
+/// let mut bpf_functions = std::collections::BTreeMap::new();
+/// register_bpf_function(&mut bpf_functions, 0, "entrypoint").unwrap();
+/// let mut executable = Executable::<UserError, DefaultInstructionMeter>::from_text_bytes(prog, bpf_functions, None, Config::default()).unwrap();
 /// let mut vm = EbpfVm::<UserError, DefaultInstructionMeter>::new(executable.as_ref(), mem, &[]).unwrap();
 ///
 /// // Provide a reference to the packet data.
@@ -391,14 +447,16 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// # Examples
     ///
     /// ```
-    /// use solana_rbpf::{vm::{Config, Executable, EbpfVm, DefaultInstructionMeter}, user_error::UserError};
+    /// use solana_rbpf::{ebpf, elf::register_bpf_function, vm::{Config, Executable, EbpfVm, DefaultInstructionMeter}, user_error::UserError};
     ///
     /// let prog = &[
     ///     0x95, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00  // exit
     /// ];
     ///
     /// // Instantiate a VM.
-    /// let executable = Executable::<UserError, DefaultInstructionMeter>::from_text_bytes(prog, None, Config::default()).unwrap();
+    /// let mut bpf_functions = std::collections::BTreeMap::new();
+    /// register_bpf_function(&mut bpf_functions, 0, "entrypoint").unwrap();
+    /// let mut executable = Executable::<UserError, DefaultInstructionMeter>::from_text_bytes(prog, bpf_functions, None, Config::default()).unwrap();
     /// let mut vm = EbpfVm::<UserError, DefaultInstructionMeter>::new(executable.as_ref(), &mut [], &[]).unwrap();
     /// ```
     pub fn new(
@@ -481,7 +539,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// # Examples
     ///
     /// ```
-    /// use solana_rbpf::{vm::{Config, Executable, EbpfVm, SyscallObject, SyscallRegistry, DefaultInstructionMeter}, syscalls::BpfTracePrintf, user_error::UserError};
+    /// use solana_rbpf::{ebpf, elf::register_bpf_function, vm::{Config, Executable, EbpfVm, SyscallObject, SyscallRegistry, DefaultInstructionMeter}, syscalls::BpfTracePrintf, user_error::UserError};
     ///
     /// // This program was compiled with clang, from a C program containing the following single
     /// // instruction: `return bpf_trace_printk("foo %c %c %c\n", 10, 1, 2, 3);`
@@ -504,7 +562,9 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// let mut syscall_registry = SyscallRegistry::default();
     /// syscall_registry.register_syscall_by_hash(6, BpfTracePrintf::call).unwrap();
     /// // Instantiate an Executable and VM
-    /// let mut executable = Executable::<UserError, DefaultInstructionMeter>::from_text_bytes(prog, None, Config::default()).unwrap();
+    /// let mut bpf_functions = std::collections::BTreeMap::new();
+    /// register_bpf_function(&mut bpf_functions, 0, "entrypoint").unwrap();
+    /// let mut executable = Executable::<UserError, DefaultInstructionMeter>::from_text_bytes(prog, bpf_functions, None, Config::default()).unwrap();
     /// executable.set_syscall_registry(syscall_registry);
     /// let mut vm = EbpfVm::<UserError, DefaultInstructionMeter>::new(executable.as_ref(), &mut [], &[]).unwrap();
     /// // Bind a context object instance to the previously registered syscall
@@ -557,7 +617,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// # Examples
     ///
     /// ```
-    /// use solana_rbpf::{vm::{Config, Executable, EbpfVm, DefaultInstructionMeter}, user_error::UserError};
+    /// use solana_rbpf::{ebpf, elf::register_bpf_function, vm::{Config, Executable, EbpfVm, DefaultInstructionMeter}, user_error::UserError};
     ///
     /// let prog = &[
     ///     0x95, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00  // exit
@@ -567,7 +627,9 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// ];
     ///
     /// // Instantiate a VM.
-    /// let executable = Executable::<UserError, DefaultInstructionMeter>::from_text_bytes(prog, None, Config::default()).unwrap();
+    /// let mut bpf_functions = std::collections::BTreeMap::new();
+    /// register_bpf_function(&mut bpf_functions, 0, "entrypoint").unwrap();
+    /// let mut executable = Executable::<UserError, DefaultInstructionMeter>::from_text_bytes(prog, bpf_functions, None, Config::default()).unwrap();
     /// let mut vm = EbpfVm::<UserError, DefaultInstructionMeter>::new(executable.as_ref(), mem, &[]).unwrap();
     ///
     /// // Provide a reference to the packet data.
@@ -616,7 +678,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
         while next_pc * ebpf::INSN_SIZE + ebpf::INSN_SIZE <= self.program.len() {
             let pc = next_pc;
             next_pc += 1;
-            let insn = ebpf::get_insn_unchecked(self.program, pc);
+            let mut insn = ebpf::get_insn_unchecked(self.program, pc);
             let dst = insn.dst as usize;
             let src = insn.src as usize;
             self.last_insn_count += 1;
@@ -675,9 +737,9 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                 },
 
                 ebpf::LD_DW_IMM  => {
-                    let next_insn = ebpf::get_insn(self.program, next_pc);
+                    ebpf::augment_lddw_unchecked(self.program, &mut insn);
                     next_pc += 1;
-                    reg[dst] = (insn.imm as u32) as u64 + ((next_insn.imm as u64) << 32);
+                    reg[dst] = insn.imm as u64;
                 },
 
                 // BPF_LDX class
@@ -747,11 +809,11 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                 },
 
                 // BPF_ALU class
-                ebpf::ADD32_IMM  => reg[dst] = (reg[dst] as i32).wrapping_add(insn.imm)          as u64,
+                ebpf::ADD32_IMM  => reg[dst] = (reg[dst] as i32).wrapping_add(insn.imm as i32)   as u64,
                 ebpf::ADD32_REG  => reg[dst] = (reg[dst] as i32).wrapping_add(reg[src] as i32)   as u64,
-                ebpf::SUB32_IMM  => reg[dst] = (reg[dst] as i32).wrapping_sub(insn.imm)          as u64,
+                ebpf::SUB32_IMM  => reg[dst] = (reg[dst] as i32).wrapping_sub(insn.imm as i32)   as u64,
                 ebpf::SUB32_REG  => reg[dst] = (reg[dst] as i32).wrapping_sub(reg[src] as i32)   as u64,
-                ebpf::MUL32_IMM  => reg[dst] = (reg[dst] as i32).wrapping_mul(insn.imm)          as u64,
+                ebpf::MUL32_IMM  => reg[dst] = (reg[dst] as i32).wrapping_mul(insn.imm as i32)   as u64,
                 ebpf::MUL32_REG  => reg[dst] = (reg[dst] as i32).wrapping_mul(reg[src] as i32)   as u64,
                 ebpf::DIV32_IMM  => reg[dst] = (reg[dst] as u32 / insn.imm as u32)               as u64,
                 ebpf::DIV32_REG  => {
@@ -906,7 +968,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                                 ..ebpf::FIRST_SCRATCH_REG + ebpf::SCRATCH_REGS],
                             next_pc,
                         )?;
-                        next_pc = self.check_pc(pc, *target_pc)?;
+                        next_pc = self.check_pc(pc, target_pc)?;
                     } else {
                         self.executable.report_unresolved_symbol(pc)?;
                     }
